# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import colour
from PySide6.QtGui import QColor


def milliseconds_to_formatted_string(milliseconds):
    """
    Converts milliseconds to a string in the format hh:mm:ss.ssss.
    """

    # Convert milliseconds to seconds
    total_seconds = milliseconds / 1000

    # Extract hours, minutes, seconds
    hours, remainder = divmod(total_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    # Format time string with leading zeros
    time_string = f"{int(hours):02d}:{int(minutes):02d}:{int(seconds):02d}"

    # Extract milliseconds (avoiding floating-point rounding issues)
    milliseconds = milliseconds % 1000
    millisecond_string = f"{milliseconds:03d}"  # Pad with leading zeros

    return (
        '<pre style="font-family: monospace">'
        f"{time_string}.{millisecond_string}"
        "</pre>"
    )


def color_fg_from_bg(bg):
    if colour.Color(bg.name()).luminance < 0.5:
        return QColor("white")
    else:
        return QColor("black")



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import os
import appdirs
from yaml import load, Loader, safe_dump
from pathlib import Path

from .constants import (
    CONFIG_FILENAME,
    APP_NAME,
)


class Config(dict):
    def __init__(self, config_file=CONFIG_FILENAME):
        self.filename = config_file
        self._load()

    def __setitem__(self, key, value):
        super().__setitem__(key, value)

    def _get_config_paths(self):
        return [
            Path(__file__).parent.joinpath("config", CONFIG_FILENAME),
            Path(appdirs.site_data_dir(APP_NAME)).joinpath(CONFIG_FILENAME),
            Path(appdirs.site_config_dir(APP_NAME)).joinpath(CONFIG_FILENAME),
            Path(appdirs.user_config_dir(APP_NAME)).joinpath(CONFIG_FILENAME),
            self.filename,
        ]

    def _load(self):
        for path in self._get_config_paths():
            config = self._load_file(path)
            if self._load_file(path):
                for k, v in config.items():
                    self.__setitem__(k, v)

    def _load_file(self, path):
        if os.path.isfile(path):
            with open(path, "r") as fid:
                return load(fid, Loader=Loader)
        else:
            return None

    def save(self):
        with open(self.filename, "w") as fid:
            safe_dump(dict(self), fid)
        pass

    def update_format(self, file_format, current_format):
        for fmt in range(file_format, current_format):
            if fmt == 1:
                pass
            if fmt == 2:
                if "timelines" in self:
                    tmp = self["timelines"]
                    del self["timelines"]
                    self["timelines"] = dict()
                    order = 0
                    for elem in sorted(tmp, key=lambda x: x["name"].casefold()):
                        name = elem["name"]
                        order += 1
                        timeline = dict()
                        timeline["order"] = order
                        if "events" in elem:
                            events = dict()
                            for ev in elem["events"]:
                                ev_dict = dict()
                                if "color" in ev:
                                    ev_dict["color"] = ev["color"]
                                events[ev["name"]] = ev_dict
                            timeline["events"] = events
                        self["timelines"][name] = timeline



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

from math import floor
from pathlib import Path
from functools import partial

from PySide6.QtCore import (
    Qt,
    QSize,
    QTimer,
)
from PySide6.QtGui import QIcon
from PySide6.QtMultimedia import (
    QAudioOutput,
    QMediaMetaData,
    QMediaPlayer,
)
from PySide6.QtMultimediaWidgets import QVideoWidget
from PySide6.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSlider,
    QStyle,
    QVBoxLayout,
    QWidget,
)

from .assets import Assets
from .constants import TIMESTAMP_HEIGHT
from .utils import (
    milliseconds_to_formatted_string,
)

with open(Path(__file__).parent.joinpath("images.py")) as f:
    exec(f.read())


class Video(QWidget):
    """A simple Media Player using Qt"""

    def __init__(self, window):
        super().__init__()
        self.media_player = QMediaPlayer()
        self.media = None
        self.mfps = None
        self.window = window

        self.create_ui()

        self.setup_timer()

    def create_ui(self):
        """Set up the user interface, signals & slots"""
        # Create the video widget
        video_widget = QVideoWidget()

        # Create Assets object
        assets = Assets()

        # Create the button layout
        self.button_box = QHBoxLayout()

        # Create the timestamp
        self.timestamp = Timestamp(self.media_player)
        self.button_box.addWidget(self.timestamp)

        self.button_box.addStretch(1)

        # Create the -10 frame button
        backward_10_frames = self.add_player_button(
            QIcon(assets.get("minus10.png")),
            "10th Previous Frame\n[Ctrl+Shift+Left]",
            partial(self.move_to_frame, -10),
        )

        # Create the -5 frame button
        backward_5_frames = self.add_player_button(
            QIcon(assets.get("minus5.png")),
            "5th Previous Frame\n[Ctrl+Left]",
            partial(self.move_to_frame, -5),
        )

        # Create the previous frame button
        previous_frame = self.add_player_button(
            QIcon(assets.get("minus1.png")),
            "Previous Frame\n[Left]",
            partial(self.move_to_frame, -1),
        )

        # Create the play/pause button
        self.play_button = self.add_player_button(
            self.style().standardIcon(QStyle.StandardPixmap.SP_MediaPlay),
            "Play\n[Space]",
            self.play_pause,
        )

        # Create the stop button
        self.stop_button = self.add_player_button(
            self.style().standardIcon(QStyle.StandardPixmap.SP_MediaStop),
            "Stop\n[s]",
            self.stop,
        )

        # Create the next frame button
        next_frame = self.add_player_button(
            QIcon(assets.get("plus1.png")),
            "Next Frame\n[Right]",
            partial(self.move_to_frame, 1),
        )

        # Create the +5 frame button
        forward_5_frames = self.add_player_button(
            QIcon(assets.get("plus5.png")),
            "5th Next Frame\n[Ctrl+Right]",
            partial(self.move_to_frame, 5),
        )

        # Create the +10 frame button
        forward_10_frames = self.add_player_button(
            QIcon(assets.get("plus10.png")),
            "10th Next Frame\n[Ctrl+Shift+Right]",
            partial(self.move_to_frame, 10),
        )

        self.frame_buttons = [
            backward_10_frames,
            backward_5_frames,
            previous_frame,
            next_frame,
            forward_5_frames,
            forward_10_frames,
        ]

        self.button_box.addStretch(1)

        # Create the volume slider
        self.volume = VolumeSlider(self)
        # Add the volume slider to the button layout
        self.button_box.addWidget(self.volume)

        # Create the main layout and add the button layout and video widget
        self.layout = QVBoxLayout()
        self.layout.addWidget(video_widget)
        self.layout.addLayout(self.button_box)

        # Setup the media player
        self.media_player.setVideoOutput(video_widget)
        self.media_player.playbackStateChanged.connect(
            self.playback_state_changed
        )
        self.media_player.mediaStatusChanged.connect(self.media_status_changed)
        self.media_player.positionChanged.connect(self.position_changed)
        self.media_player.durationChanged.connect(self.duration_changed)

        # Setup the audio output
        self.audio_output = QAudioOutput()
        self.media_player.setAudioOutput(self.audio_output)

        # Prevent the individual UIs from getting the focus
        for ui in [
            self.play_button,
            self.stop_button,
            self.volume,
            video_widget,
        ] + self.frame_buttons:
            ui.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        # Add the main layout to the main window
        self.setLayout(self.layout)

    def setup_timer(self):
        self.timer = QTimer(self)
        self.timer.setInterval(100)
        self.timer.timeout.connect(
            partial(self.timestamp.update, self.media_player.position())
        )

    def play_pause(self):
        """Toggle play/pause status"""
        if (
            self.media_player.playbackState()
            == QMediaPlayer.PlaybackState.PlayingState
        ):
            self.media_player.pause()
        else:
            self.media_player.play()

    def stop(self):
        """Stop player"""
        self.media_player.stop()
        self.play_button.setIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_MediaPlay)
        )

    def playback_state_changed(self, state):
        """Set the button icon when media changes state"""
        if state == QMediaPlayer.PlaybackState.PlayingState:
            self.timer.start()
            self.play_button.setIcon(
                self.style().standardIcon(QStyle.StandardPixmap.SP_MediaPause)
            )
        else:
            self.timer.stop()
            self.play_button.setIcon(
                self.style().standardIcon(QStyle.StandardPixmap.SP_MediaPlay)
            )
            # FIXME: Stop at the correct frame
            self.set_position(self.media_player.position())

        self.stop_button.setEnabled(
            state != QMediaPlayer.PlaybackState.StoppedState
        )
        self.window.menu.stop_action.setEnabled(
            state != QMediaPlayer.PlaybackState.StoppedState
        )
        for ui in self.frame_buttons:
            ui.setEnabled(state == QMediaPlayer.PlaybackState.PausedState)

        for action in [
            self.window.menu.previous_frame_action,
            self.window.menu.next_frame_action,
            self.window.menu.fifth_previous_frame_action,
            self.window.menu.tenth_previous_frame_action,
            self.window.menu.fifth_next_frame_action,
            self.window.menu.tenth_next_frame_action,
            self.window.menu.add_timeline_action,
            self.window.menu.add_occurrence_action,
        ]:
            action.setEnabled(state == QMediaPlayer.PlaybackState.PausedState)

    def media_status_changed(self, state):
        if state == QMediaPlayer.MediaStatus.LoadedMedia:
            # Enable play button
            self.window.menu.play_action.setEnabled(True)
            # Enable save project button
            self.window.menu.save_project_action.setEnabled(True)
            # Check if metadata is available
            metadata = self.media_player.metaData()
            if metadata:
                # If metadata is available, set the frame rate
                fps = metadata.value(QMediaMetaData.Key.VideoFrameRate)
                self.mfps = int(1000 / fps)
            else:
                self.mfps = None

        if state == QMediaPlayer.MediaStatus.BufferedMedia:
            self.window.media_status_changed()

    def position_changed(self, position):
        """Update the position slider"""
        self.timestamp.update(self.media_player.position())
        self.window.position_changed(position)

    def duration_changed(self, duration):
        """Update the duration slider"""
        self.window.duration_changed(duration)

    def set_position(self, position):
        """Set the position"""
        position = int(self.mfps * floor(position / self.mfps) + self.mfps / 2)
        if position < 0:
            position = int(self.mfps / 2)
        self.window.set_position(position)

    # The argument "checked" is necessary, in order to have this function used
    # via functools.partial in the triggered option of QAction
    def move_to_frame(self, nb_frame, checked=False):
        state = self.media_player.playbackState()
        if self.mfps is None or state != QMediaPlayer.PlaybackState.PausedState:
            return
        self.set_position(self.media_player.position() + (self.mfps * nb_frame))

    def add_player_button(self, icon, tooltip, cbfunc):
        ui = QPushButton()
        ui.setEnabled(False)
        ui.setFixedHeight(24)
        ui.setIconSize(QSize(16, 16))
        ui.setIcon(icon)
        ui.setToolTip(tooltip)
        ui.clicked.connect(cbfunc)
        self.button_box.addWidget(ui)
        return ui

    def set_source(self, media):
        self.media = media

        # Set the source of the media player
        self.media_player.setSource(media)

        # Show first frame
        self.media_player.play()
        self.media_player.pause()

        # Enable the buttons
        self.play_button.setEnabled(True)
        self.stop_button.setEnabled(True)


class Timestamp(QLabel):

    def __init__(self, media_player):
        super().__init__()
        self.set_text(0)
        self.setFixedHeight(TIMESTAMP_HEIGHT)
        self.media_player = media_player

    def set_text(self, time):
        self.setText(milliseconds_to_formatted_string(time))

    def update(self, time):
        """Update the timestamp"""
        self.set_text(time)


class VolumeSlider(QSlider):

    def __init__(self, parent):
        super().__init__(Qt.Orientation.Horizontal, parent)
        self.parent = parent
        self.setRange(0, 100)
        self.setValue(100)
        self.valueChanged.connect(self.set_audio_volume)
        self.update_tooltip()

    def set_audio_volume(self, volume):
        self.parent.audio_output.setVolume(volume / 100)
        self.update_tooltip()

    def update_tooltip(self):
        self.setToolTip(f"Audio volume ({self.value()}%)")



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QTextEdit


class TextEdit(QTextEdit):
    def __init__(self, parent, text):
        super().__init__()
        self.append(text)
        self.parent = parent

    def keyPressEvent(self, event):
        if (
            event.key() == Qt.Key.Key_Return
            and event.modifiers() == Qt.KeyboardModifier.NoModifier
        ):
            self.parent.accept()
        else:
            if event.key() == Qt.Key.Key_Return:
                self.append("")
            else:
                super().keyPressEvent(event)



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

from PySide6.QtWidgets import QMessageBox

FORMAT = 3


def format_ok(fmt, widget=None):
    msg = QMessageBox(widget)
    msg.setWindowTitle("ViCodePy: Project file format issue")

    if not isinstance(fmt, int):
        msg.setIcon(QMessageBox.Critical)
        msg.setText("The format of the project file must be a integer number")
        msg.exec()
        return False

    if fmt < FORMAT:
        msg.setIcon(QMessageBox.Warning)
        msg.setText("The project file format is outdated.")
        msg.setInformativeText(
            f"The format version indicated in the project file ({fmt}) "
            "is outdated in respect to the format version supported by "
            f"this version of ViCodePy ({FORMAT}).\n\n"
            "The project file will be upgraded accordingly."
        )
        msg.setStandardButtons(QMessageBox.Ok)
        msg.exec()
        return True

    if fmt > FORMAT:
        msg.setIcon(QMessageBox.Critical)
        msg.setText("The current installed version of ViCodePy is outdated.")
        msg.setInformativeText(
            "The format version supported by this version of ViCodePy "
            f"({FORMAT}) is outdated in respect to the format version "
            f"indicated in the project file ({fmt}).\n\n"
            "Please, upgrade ViCodePy to the last avaiable version."
        )
        msg.setStandardButtons(QMessageBox.Ok)
        msg.exec()
        return False

    return True



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import importlib.metadata
from pathlib import Path
from PySide6.QtWidgets import QMessageBox
from PySide6.QtCore import QUrl
from PySide6.QtGui import QDesktopServices

from .format import FORMAT
from .constants import (
    APP_NAME,
    REPOSITORY_URL,
    PYPI_URL,
)


def open_repository_url():
    open_url(REPOSITORY_URL)


def open_pypi_url():
    open_url(PYPI_URL)


def open_url(url):
    qurl = QUrl(url)
    if not QDesktopServices.openUrl(qurl):
        QMessageBox.warning("Open Url", f"Could not open url '{url}'")


class About:

    def get_pkg_version(self) -> str:
        """Find the version of this package."""
        try:

            pyproject_toml_file = (
                Path(__file__).parent.parent / "pyproject.toml"
            )

            with open(pyproject_toml_file) as fid:
                import tomlkit

                toml_content = tomlkit.parse(fid.read())
                version = f"{toml_content['project']['version']} (dev)"

        except FileNotFoundError:

            try:

                version = importlib.metadata.version(APP_NAME)

            except importlib.metadata.PackageNotFoundError:

                version = "not found"

        return version

    def get_fmt_version(self) -> str:
        """Find the version of the format of the project file"""
        try:

            with open(Path(__file__).parent.joinpath("format.py")) as f:
                exec(f.read())

        except FileNotFoundError:
            pass

        return FORMAT

    def exec(self):
        msg = QMessageBox()
        msg.setWindowTitle("About")
        msg.setText(
            "<center><font size='+1'><b>ViCodePy</b></font></center><br/><br/>"
            "Copyright © 2024 Rafael Laboissière<br/>"
            "Copyright © 2024 Esteban Milleret<br/>"
            "Licensed under the terms of the <a href='https://www.gnu.org/"
            "licenses/gpl-3.0.en.html'>GPL v3 or later</a><br/>"
            f"<a href='{PYPI_URL}'>PyPI project</a><br/>"
            f"<a href='{REPOSITORY_URL}'>Git repository</a><br/>"
        )
        msg.setInformativeText(
            f"Version {self.get_pkg_version()}\n"
            f"Project file format version {self.get_fmt_version()}"
        )
        msg.setStandardButtons(QMessageBox.Ok)
        msg.exec()



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import os

from PySide6.QtCore import (
    QEvent,
    QSize,
    Qt,
)
from PySide6.QtMultimedia import QMediaPlayer
from PySide6.QtWidgets import (
    QMainWindow,
    QMessageBox,
    QSplitter,
)

from .assets import Assets
from .constants import (
    MAIN_WINDOW_MINIMUM_WIDTH,
    MAIN_WINDOW_MINIMUM_HEIGHT,
)
from .files import Files
from .menu import Menu
from .video import Video
from .timepane import TimePaneView


class MainWindow(QMainWindow):

    def __init__(self):
        super().__init__()
        self.files = Files(self)
        self.files.data_to_load = None

        self.setWindowTitle("ViCodePy")
        self.setMinimumSize(
            QSize(
                MAIN_WINDOW_MINIMUM_WIDTH,
                MAIN_WINDOW_MINIMUM_HEIGHT,
            )
        )

        # Load the QSS file
        with open(Assets().get("style.qss"), "r") as f:
            qss = f.read()
        self.setStyleSheet(qss)

        self.installEventFilter(self)

        self.create_ui()

    def create_ui(self):
        """Set up the user interface, signals & slots"""
        self.video = Video(self)

        splitter = QSplitter(Qt.Orientation.Vertical)
        self.setCentralWidget(splitter)
        splitter.addWidget(self.video)

        # Add TimePaneView
        view = TimePaneView(self)
        splitter.addWidget(view)

        self.time_pane = view.scene.time_pane
        self.scene = view.scene

        self.menu = Menu(self)

    def on_fullscreen(self):
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    def closeEvent(self, event):
        """Display a warning dialog to user when app is closing"""

        action = "ok"
        if self.time_pane.data_needs_save:
            msgBox = QMessageBox(
                QMessageBox.Icon.Warning,
                "Quit the application",
                (
                    "You are about to quit the application. "
                    "The changes made in this session will be lost."
                ),
                QMessageBox.StandardButton.Cancel
                | QMessageBox.StandardButton.Save
                | QMessageBox.StandardButton.Ok,
                self,
            )

            msgBox.button(QMessageBox.StandardButton.Save).setText(
                "Save and Quit"
            )
            msgBox.button(QMessageBox.StandardButton.Ok).setText("Quit")
            msgBox.exec()

            if msgBox.clickedButton() == msgBox.button(
                QMessageBox.StandardButton.Ok
            ):
                action = "ok"
            elif msgBox.clickedButton() == msgBox.button(
                QMessageBox.StandardButton.Save
            ):
                action = "save"
            else:
                action = "ignore"

        if action == "ok":
            # Clean up temp dir
            self.files.temp_dir_cleanup()
            event.accept()
        elif action == "save":
            if self.files.save_project():
                # Clean up temp dir
                self.files.temp_dir_cleanup()
                event.accept()
            else:
                event.ignore()
        else:
            event.ignore()

    def eventFilter(self, obj, event):
        if event.type() == QEvent.Type.KeyRelease:
            if event.key() == Qt.Key.Key_Right:
                if event.isAutoRepeat():
                    if (
                        self.video.media_player.playbackState()
                        != QMediaPlayer.PlaybackState.PlayingState
                    ):
                        self.video.media_player.play()
                else:
                    self.video.media_player.pause()
                return True
        return False

    def media_status_changed(self):
        if self.files.data_to_load:
            self.files.load_data_file(self.files.data_to_load)
            self.time_pane.changed = False

    def position_changed(self, position):
        self.time_pane.time = position
        self.time_pane.view.update()

    def duration_changed(self, duration):
        self.time_pane.duration = duration
        self.time_pane.view.update()

    def set_position(self, position):
        self.time_pane.view.set_position(position)

    def load_file(self, path):
        self.files.load_file(path)

    def no_video_loaded(self):
        self.files.no_video_loaded()

    def set_source(self, media):
        # Change video player media
        self.video.set_source(media)

        # Set the media filename as window title
        self.setWindowTitle(os.path.basename(media.path()))

        # Clear the time_pane
        self.time_pane.clear()
        self.scene.create_time_pane()
        self.time_pane = self.scene.time_pane
        self.time_pane.view.update_scale()

    def has_media(self):
        return self.video.media

    def set_video_position(self, time):
        self.video.set_position(time)

    def set_media_player_position(self, time):
        self.video.media_player.setPosition(int(time))

    def media_player_playback_state(self):
        self.video.media_player.playbackState()

    def media_player_pause(self):
        self.video.media_player.pause()

    def mfps(self):
        return self.video.mfps



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.


from PySide6.QtCore import (
    QRectF,
    QLine,
    QPointF,
    QSizeF,
)
from PySide6.QtGui import QPolygonF
from PySide6.QtWidgets import QGraphicsItem

from .constants import (
    CURSOR_BRUSH_COLOR,
    CURSOR_HANDLE_WIDTH,
    CURSOR_PEN_COLOR,
    TIME_SCALE_HEIGHT,
)


class Cursor(QGraphicsItem):

    def __init__(self, parent):
        super().__init__(parent)
        self.time_pane = parent.time_pane
        self.pressed = False
        self.poly: QPolygonF = QPolygonF(
            [
                QPointF(-CURSOR_HANDLE_WIDTH, 0),
                QPointF(CURSOR_HANDLE_WIDTH, 0),
                QPointF(0, TIME_SCALE_HEIGHT),
            ]
        )
        # The constant value 10000 below should be enough, in practical terms
        self.line: QLine = QLine(0, TIME_SCALE_HEIGHT, 0, 10000)

        self.setAcceptHoverEvents(True)
        self.setAcceptDrops(True)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsFocusable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges)
        self.setZValue(101)

    def paint(self, painter, option, widget=...):
        painter.setPen(CURSOR_PEN_COLOR)
        painter.setBrush(CURSOR_BRUSH_COLOR)
        painter.drawLine(self.line)
        painter.drawPolygon(self.poly)

    def calculate_size(self):
        min_x: float = self.poly[0].x()
        max_x: float = self.poly[0].x()

        for i, point in enumerate(self.poly):
            if point.x() < min_x:
                min_x = point.x()
            if point.x() > max_x:
                max_x = point.x()

        return QSizeF(max_x - min_x, TIME_SCALE_HEIGHT)

    def boundingRect(self):
        size: QSizeF = self.calculate_size()
        return QRectF(-CURSOR_HANDLE_WIDTH, 0, size.width(), size.height())

    def focusInEvent(self, event):
        self.pressed = True
        super().focusInEvent(event)
        self.update()

    def focusOutEvent(self, event):
        self.pressed = False
        super().focusOutEvent(event)
        self.update()

    def mouseMoveEvent(self, event):
        super().mouseMoveEvent(event)
        pos: QPointF = event.scenePos()
        if self.pressed:
            time = int(
                pos.x()
                * self.time_pane.duration
                / self.time_pane.scene.width()
            )

            # During creation of a new occurrence
            if self.time_pane and self.time_pane.occurrence_under_creation:
                occurrence = self.time_pane.occurrence_under_creation
                if time != occurrence.get_time_from_bounding_interval(time):
                    # Stop player at the lower or upper bound when they
                    # are passed over
                    self.setPos(self.x(), 0)
                    return

            self.time_pane.window.set_video_position(time)

            if pos.x() < 0:
                self.setPos(0, 0)
            elif pos.x() > self.time_pane.scene.width():
                self.setPos(self.time_pane.scene.width(), 0)
            else:
                self.setPos(pos.x(), 0)

        self.update()



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import pandas as pd
from math import inf

from PySide6.QtCore import Qt
from PySide6.QtGui import (
    QPainter,
    QColor,
)
from PySide6.QtMultimedia import QMediaPlayer
from PySide6.QtWidgets import (
    QAbstractSlider,
    QGraphicsRectItem,
    QGraphicsScene,
    QGraphicsView,
    QMessageBox,
    QScrollBar,
    QStyle,
)
from .constants import (
    CSV_HEADERS,
    EVENT_DEFAULT_COLOR,
    MINIMUM_ZOOM_FACTOR,
    OCCURRENCE_PEN_WIDTH_OFF_CURSOR,
    OCCURRENCE_PEN_WIDTH_ON_CURSOR,
    TIMELINE_HEIGHT,
    TIME_SCALE_HEIGHT,
    ZOOM_STEP,
)
from .dialog import (
    ConfirmMessageBox,
    DialogCode,
    TimelineDialog,
)
from .event import (
    ChooseEvent,
    Event,
    EventCollection,
)
from .occurrence import (
    Occurrence,
    OccurrenceHandle,
)
from .timeline import Timeline
from .timescale import TimeScale


class TimePaneView(QGraphicsView):

    def __init__(self, window):
        super().__init__(window)
        self.zoom_step = ZOOM_STEP
        self.zoom_shift = None
        self.minimum_zoom_factor = MINIMUM_ZOOM_FACTOR
        self.zoom_factor = self.minimum_zoom_factor
        self.window = window
        self.create_ui()

    def create_ui(self):
        vertical_scrollbar = QScrollBar(Qt.Orientation.Vertical, self)
        vertical_scrollbar.valueChanged.connect(
            self.on_vertical_scroll_value_changed
        )
        self.setVerticalScrollBar(vertical_scrollbar)

        self.setRenderHint(QPainter.RenderHint.Antialiasing)
        self.setFrameShape(QGraphicsView.Shape.NoFrame)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        # This is necessary for getting the cursor being updated
        self.setViewportUpdateMode(
            QGraphicsView.ViewportUpdateMode.FullViewportUpdate
        )
        self.scene = TimePaneScene(self)
        self.scene.setSceneRect(0, 0, self.width(), self.height())
        self.setScene(self.scene)

    def wheelEvent(self, event):
        if event.modifiers() == Qt.KeyboardModifier.ControlModifier:
            if not self.window.has_media():
                return
            mouse_pos = self.mapToScene(event.position().toPoint()).x()
            if event.angleDelta().y() > 0:
                self.zoom_shift = mouse_pos * (1 - self.zoom_step)
                self.zoom_in()
            else:
                self.zoom_shift = mouse_pos * (1 - 1 / self.zoom_step)
                self.zoom_out()
            self.zoom_shift = None
        elif event.modifiers() == Qt.KeyboardModifier.ShiftModifier:
            if event.angleDelta().y() > 0:
                action = QAbstractSlider.SliderSingleStepAdd
            else:
                action = QAbstractSlider.SliderSingleStepSub
            self.horizontalScrollBar().triggerAction(action)
        else:
            super().wheelEvent(event)

    def on_vertical_scroll_value_changed(self, value):
        """Allow the time scale to be always visible when scrolling"""
        if self.scene.time_pane.time_scale:
            self.scene.time_pane.time_scale.setPos(0, value)

    def zoom_in(self):
        self.zoom_factor *= self.zoom_step
        self.update_scale()

    def zoom_out(self):
        new_zoom_factor = self.zoom_factor / self.zoom_step
        if new_zoom_factor >= self.minimum_zoom_factor:
            self.zoom_factor = new_zoom_factor
            self.update_scale()

    def update_scale(self):
        # Update the size of the scene with zoom_factor
        self.scene.setSceneRect(
            0,
            0,
            self.width() * self.zoom_factor,
            self.scene.height(),
        )

        if self.zoom_shift:
            previous_anchor = self.transformationAnchor()
            self.setTransformationAnchor(QGraphicsView.NoAnchor)
            self.translate(self.zoom_shift, 0)
            self.setTransformationAnchor(previous_anchor)

    def mouseDoubleClickEvent(self, event):
        super().mouseDoubleClickEvent(event)

        # Get the position click from the scene
        map = self.mapToScene(self.scene.sceneRect().toRect())
        x = map.boundingRect().x()

        # Compute the time of the position click
        time = int(
            (x + event.scenePosition().x())
            * self.scene.time_pane.duration
            / self.scene.width()
        )

        self.window.set_video_position(time)

    def set_position(self, time):
        # During the creation of a new occurrence
        if self.scene.time_pane.occurrence_under_creation:
            occurence = self.scene.time_pane.occurrence_under_creation
            time = occurence.get_time_from_bounding_interval(time)

        # Cope with selected occurrence
        for i in self.scene.time_pane.scene.selectedItems():
            if isinstance(i, Occurrence):
                time = i.get_time_from_bounding_interval(time)
                break

        # Set time of the video player
        self.window.set_media_player_position(int(time))

    def resizeEvent(self, a0):
        # FIXME: Instead of testing the presence of the time scale, test whether there is a video loaded
        if self.scene.time_pane.time_scale:
            origin = self.mapToScene(0, 0).x()
            width_before = self.scene.width() / self.zoom_factor
            width_after = self.width()
            shift = origin * (1 - width_after / width_before)
            self.update_scale()
            previous_anchor = self.transformationAnchor()
            self.setTransformationAnchor(QGraphicsView.NoAnchor)
            self.translate(shift, 0)
            self.setTransformationAnchor(previous_anchor)
        else:
            self.scene.setSceneRect(
                0,
                0,
                self.width(),
                self.scene.height(),
            )
        self.update()
        super().resizeEvent(a0)

    def keyPressEvent(self, event):
        pass


class TimePaneScene(QGraphicsScene):

    def __init__(self, view):
        super().__init__()
        self.view = view
        self.sceneRectChanged.connect(self.on_scene_changed)
        self.selectionChanged.connect(self.on_selection_changed)
        self.create_time_pane()

    def create_time_pane(self):
        self.time_pane = TimePane(self)
        self.addItem(self.time_pane)

    def on_selection_changed(self):
        selected_items = self.selectedItems()
        selected = None
        if len(selected_items) == 1:
            selected = selected_items[0]
            if isinstance(selected, Timeline):
                self.time_pane.selected_timeline = selected
        for s in self.time_pane.timelines():
            s.select = s == selected

    def on_scene_changed(self, rect):
        self.time_pane.on_change(rect)


class TimePane(QGraphicsRectItem):

    def __init__(self, scene):
        """Initialize the time pane graphics item"""
        super().__init__()
        self._duration = 0
        self._time = 0

        self.selected_timeline = None
        self.occurrence_under_creation: Occurrence = None
        self.scene = scene
        self.view = scene.view
        self.window = self.view.window
        self.time_scale = None
        self.data_needs_save = False
        self.scrollbar_width = self.window.style().pixelMetric(
            QStyle.PM_ScrollBarExtent
        )

    def on_change(self, rect):
        # Update occurrences
        for timeline in self.timelines():
            timeline.update_rect_width(rect.width())
            for occurrence in timeline.occurrences:
                occurrence.update_rect()

        if self.occurrence_under_creation:
            self.occurrence_under_creation.update_rect()

        # Update time scale display
        if self.time_scale:
            # Update cursor
            if self.duration:
                self.time_scale.cursor.setX(self.time * rect.width() / self.duration)
            self.time_scale.update_rect()

    def select_cycle_timeline(self, delta):
        timelines = self.timelines()
        i, n = self.find_selected_timeline()
        selected_timeline = timelines[i]
        selected_timeline.select = False
        if delta > 0:
            if i == n - 1:
                i = -1
        else:
            if i == 0:
                i = n
        i += delta
        self.select_timeline(timelines[i])

    def find_selected_timeline(self):
        timelines = self.timelines()
        n = len(timelines)
        for i in range(n):
            if timelines[i].select:
                break
        return i, n

    def select_timeline(self, line):
        for tl in self.timelines():
            tl.select = False
        line.select = True
        self.selected_timeline = line
        line.update()

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, time):
        if time != self._time:
            self._time = time
            self.on_time_changed(time)

    def on_time_changed(self, new_time):
        # First, update the current occurrence, if it exists. If the cursor
        # time goes beyond the allowed bounds, bring it back and do not update
        # the other widgets.
        if self.occurrence_under_creation:
            if (
                self.occurrence_under_creation.lower_bound
                and new_time < self.occurrence_under_creation.lower_bound
            ):
                new_time = self.occurrence_under_creation.lower_bound
            elif (
                self.occurrence_under_creation.upper_bound
                and new_time > self.occurrence_under_creation.upper_bound
            ):
                new_time = self.occurrence_under_creation.upper_bound
                if (
                    self.window.media_player_playback_state()
                    == QMediaPlayer.PlaybackState.PlayingState
                ):
                    self.window.media_player_pause()
            begin_time = self.occurrence_under_creation.begin_time
            mfps = self.window.mfps()
            if new_time >= begin_time:
                self.occurrence_under_creation.update_end_time(
                    new_time + int(mfps / 2)
                )
            else:
                self.occurrence_under_creation.update_end_time(
                    new_time - int(mfps / 2)
                )

        # Cope with selected occurrence
        for i in self.scene.selectedItems():
            if isinstance(i, Occurrence):
                new_time = i.get_time_from_bounding_interval(new_time)
                break

        # Update cursor position
        if self.time_scale.cursor:
            self.time_scale.cursor.setX(new_time * self.scene.width() / self.duration)

        if isinstance(self.scene.focusItem(), OccurrenceHandle):
            occurrence_handle: OccurrenceHandle = self.scene.focusItem()
            occurrence_handle.change_time(new_time)

        # Change appearance of occurrence under the cursor
        # (Brute force approach; this ought to be improved)
        if not self.occurrence_under_creation:
            for t in self.timelines():
                for a in t.occurrences:
                    a.pen_width = OCCURRENCE_PEN_WIDTH_OFF_CURSOR
            if self.selected_timeline:
                for a in self.selected_timeline.occurrences:
                    if a.begin_time <= new_time and a.end_time >= new_time:
                        a.pen_width = OCCURRENCE_PEN_WIDTH_ON_CURSOR
                        break

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, duration):
        if duration != self._duration:
            self._duration = duration
            # Recreate time pane scale and the cursor
            self.time_scale = TimeScale(self)
            self.view.update()

    def clear(self):
        # Clear timelineScene
        self.scene.clear()

    def handle_occurrence(self):
        """Handle the occurrence"""
        menu = self.window.menu
        if self.occurrence_under_creation is None:
            if self.selected_timeline.can_create_occurrence(self.time):
                mfps = self.window.mfps()
                self.occurrence_under_creation = Occurrence(
                    self.selected_timeline,
                    self.time - int(mfps / 2),
                    self.time + int(mfps / 2),
                )
            menu.finish_occurence()
        else:
            # Finish the current occurrence
            events_dialog = ChooseEvent(
                self.selected_timeline.event_collection, self.view
            )
            events_dialog.exec()
            if events_dialog.result() == QMessageBox.DialogCode.Accepted:
                event = events_dialog.get_chosen()
                self.occurrence_under_creation.set_event(event)
                self.occurrence_under_creation.finish_creation()
                self.occurrence_under_creation = None
                self.on_time_changed(self.time)
            menu.start_occurence()
            self.data_needs_save = True
            self.view.update()

    def handle_timeline(self):
        dialog = TimelineDialog(self.window)
        dialog.exec()
        if dialog.result() == DialogCode.Accepted:
            name = dialog.get_text()
            if name == "":
                QMessageBox.warning(
                    self, "Warning", "Timeline name cannot be empty"
                )
                return
            self.add_timeline(Timeline(name, self))
            self.data_needs_save = True

    def abort_occurrence_creation(self):
        if self.occurrence_under_creation is not None:
            confirm_box = ConfirmMessageBox("Abort creation of occurrence?")
            if confirm_box.result() == ConfirmMessageBox.DialogCode.Accepted:
                self.occurrence_under_creation.remove()
                self.occurrence_under_creation = None
                self.view.update()
                self.window.menu.start_occurence()

    def add_timeline(self, timeline):
        timeline.update_rect()

        # Compute the new height of the scene
        new_height = TIME_SCALE_HEIGHT + len(self.timelines()) * TIMELINE_HEIGHT
        scene_rect = self.scene.sceneRect()
        scene_rect.setHeight(new_height)
        self.scene.setSceneRect(scene_rect)

        # Set maximum height of the widget
        self.view.setMaximumHeight(int(new_height) + self.scrollbar_width + 2)

        # Select the new timeline
        for i in self.timelines():
            i.select = False
        timeline.select = True

        self.view.update_scale()

    def add_data(self, data):
        for _, row in data.data.iterrows():
            # Search for timeline
            timeline = self.get_timeline_by_name(row["timeline"])

            # If timeline from csv doesn't exist in TimePane,
            # escape row
            if not timeline:
                continue

            # Search for event
            event = timeline.event_collection.get_event(row["event"])

            # If event from csv doesn't exist in timeline,
            # then add it
            if not event:
                continue

            occurrence = Occurrence(
                timeline,
                int(row["begin"]),
                int(row["end"]),
            )

            occurrence.set_event(event)
            occurrence.finish_creation()

    def get_timeline_by_name(self, name):
        """Get the timeline by name"""
        return next((x for x in self.timelines() if x.name == name), None)

    def has_occurrences(self) -> bool:
        return any(len(line.occurrences) for line in self.timelines())

    def delete_occurrence(self):
        for i in self.scene.selectedItems():
            if isinstance(i, Occurrence):
                i.on_remove()
                break

    def timelines_from_config(self, config):
        if "timelines" in config:

            # Set all absent order fields with Inf
            for k, v in config["timelines"].items():
                if not v:
                    v = dict()
                    config["timelines"][k] = v
                if "order" not in v:
                    v["order"] = -inf
                if "description" not in v:
                    v["description"] = ""

            # Sort according to order first and alphabetically from
            # timeline name, otherwise
            for item in sorted(
                config["timelines"].items(),
                key=lambda x: (x[1]["order"], x[0]),
            ):
                # Get name and properties of the timeline
                name = item[0]
                properties = item[1]

                # Create timeline
                line = Timeline(name, self)

                # Check description
                description = properties["description"]
                if description == "":
                    description = "(no description)"
                line.title.setToolTip(description)
                line.description = description

                # Add the timeline to the TimePane
                self.add_timeline(line)

                # Loop over events of the timeline
                event_collection = EventCollection()
                if "events" in properties:
                    for k, v in properties["events"].items():
                        event = Event(k)
                        try:
                            event.color = QColor(v["color"])
                        except KeyError:
                            event.color = EVENT_DEFAULT_COLOR
                        try:
                            event.description = v["description"]
                        except KeyError:
                            event.description = ""
                        event_collection.add_event(event)
                    line.event_collection = event_collection

    def timelines_to_dataframe(self):
        df = pd.DataFrame(columns=CSV_HEADERS)
        for timeline in sorted(self.timelines(), key=lambda x: x.name):
            for occurrence in timeline.occurrences:
                comment = occurrence.comment.replace('"', '\\"')
                row = [
                    timeline.name,
                    occurrence.event.name,
                    occurrence.begin_time,
                    occurrence.end_time,
                    comment,
                ]
                df.loc[len(df.index)] = row
        return df

    def timelines(self):
        return [x for x in self.childItems() if isinstance(x, Timeline)]



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

from enum import IntEnum

from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QLineEdit,
    QMessageBox,
)


class TimelineDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Add a new timeline")

        self.name_edit = QLineEdit(self)
        layout = QFormLayout(self)
        layout.addRow("Enter name of the new timeline:", self.name_edit)

        buttons = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok
            | QDialogButtonBox.StandardButton.Cancel,
            self,
        )
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

    def get_text(self):
        return self.name_edit.text()


class ConfirmMessageBox(QMessageBox):
    def __init__(self, message, parent=None):
        super().__init__(
            QMessageBox.Icon.Warning,
            "Warning",
            message,
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            parent,
        )

        self.button(QMessageBox.StandardButton.Yes).clicked.connect(self.accept)
        self.button(QMessageBox.StandardButton.No).clicked.connect(self.reject)
        self.exec()


class DialogCode(IntEnum):
    Accepted: int = QDialog.DialogCode.Accepted  # 0
    Canceled: int = QDialog.DialogCode.Rejected  # 1
    Aborted: int = 2



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

from PySide6.QtCore import Qt
from PySide6.QtGui import QColor

# Configuration and metadada file
CONFIG_FILENAME = "config.yml"
METADATA_FILENAME = "metadata.yml"

# Application name
APP_NAME = "vicodepy"

# URLs for the About window
REPOSITORY_URL = (
    "https://gricad-gitlab.univ-grenoble-alpes.fr/babylab-lpnc/video-coder"
)
PYPI_URL = "https://pypi.org/project/vicodepy/"

# Default dimensions of the main application window
APP_WINDOW_WIDTH = 640
APP_WINDOW_HEIGHT = 480

# Time scale height
TIME_SCALE_HEIGHT = 25

# Cursor appearance
CURSOR_PEN_COLOR = Qt.GlobalColor.black
CURSOR_BRUSH_COLOR = QColor(0, 0, 0, 100)
CURSOR_HANDLE_WIDTH = 10

# Minimum size of main window
MAIN_WINDOW_MINIMUM_WIDTH = 640
MAIN_WINDOW_MINIMUM_HEIGHT = 320

# Occurrence appearance
OCCURRENCE_PEN_COLOR = QColor(0, 0, 0)
OCCURRENCE_BG_COLOR = QColor(128, 128, 128)
OCCURRENCE_PEN_WIDTH_ON_CURSOR = 3
OCCURRENCE_PEN_WIDTH_OFF_CURSOR = 1
OCCURRENCE_HANDLE_WIDTH = 9
OCCURRENCE_HANDLE_HEIGHT_FRACTION = 0.5

# Tick locator parameter
TICK_LOCATOR_MIN_GAP = 0.05

# Timeline appearance
TIMELINE_HEIGHT = 60
TIMELINE_TITLE_HEIGHT = 20
TIMELINE_TITLE_ON_BG_COLOR = QColor(40, 40, 40)
TIMELINE_TITLE_OFF_BG_COLOR = QColor(200, 200, 200)
TIMELINE_TITLE_ON_FG_COLOR = QColor(200, 200, 200)
TIMELINE_TITLE_OFF_FG_COLOR = QColor(150, 150, 150)

# Event default color
EVENT_DEFAULT_COLOR = QColor(255, 255, 255)

# Data file constants
CSV_HEADERS = ["timeline", "event", "begin", "end", "comment"]
CSV_DELIMITER = ","

# GraphicsView zooming constants
MINIMUM_ZOOM_FACTOR = 1.0
ZOOM_STEP = 1.2

# Timestamp height
TIMESTAMP_HEIGHT = 24



# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import platform
from functools import partial

from PySide6.QtCore import Qt
from PySide6.QtGui import (
    QAction,
    QKeySequence,
)
from PySide6.QtWidgets import QStyle

from .about import About, open_repository_url, open_pypi_url


class Menu:

    def __init__(self, window):

        self.window = window
        self.menu_bar = window.menuBar()
        self.style = self.window.style()

        self.create_file_menu()
        self.create_play_menu()
        self.create_edit_menu()
        self.create_view_menu()
        self.create_help_menu()

    def create_file_menu(self):
        menu = self.menu_bar.addMenu("&File")

        # Add actions to file menu
        open_action = QAction(
            self.style.standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon),
            "&Open video",
            self.window,
            shortcut=QKeySequence.StandardKey.Open,
            triggered=self.window.files.open_video,
        )

        open_project_action = QAction(
            self.style.standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon),
            "Open &project",
            self.window,
            shortcut=QKeySequence(
                Qt.Modifier.CTRL | Qt.Modifier.SHIFT | Qt.Key.Key_O
            ),
            triggered=self.window.files.open_project,
        )

        icon = self.style.standardIcon(
            QStyle.StandardPixmap.SP_DialogSaveButton
        )
        self.save_project_action = QAction(
            icon,
            "Save project",
            self.window,
            shortcut=QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_S),
            triggered=self.window.files.save_project,
            enabled=False,
        )

        export_action = QAction(
            icon,
            "Export Data",
            self.window,
            shortcut=QKeySequence(
                Qt.Modifier.CTRL | Qt.Modifier.SHIFT | Qt.Key.Key_S
            ),
            triggered=self.window.files.export_data_file,
        )

        close_action = QAction(
            "Quit",
            self.window,
            shortcut=(
                QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Q)
                if platform.system() == "Windows"
                else QKeySequence.StandardKey.Quit
            ),
            triggered=self.window.close,
        )

        menu.addAction(open_action)
        menu.addAction(open_project_action)
        menu.addAction(self.save_project_action)
        menu.addAction(export_action)
        menu.addAction(close_action)

    def create_play_menu(self):
        # Add actions to play menu
        menu = self.menu_bar.addMenu("&Play")
        icon = self.style.standardIcon(QStyle.StandardPixmap.SP_MediaPlay)
        self.play_action = QAction(
            icon,
            "Play/Pause",
            self.window,
            shortcut=Qt.Key.Key_Space,
            triggered=self.window.video.play_pause,
            enabled=False,
        )

        icon = self.style.standardIcon(QStyle.StandardPixmap.SP_MediaStop)
        self.stop_action = QAction(
            icon,
            "Stop",
            self.window,
            shortcut=Qt.Key.Key_S,
            triggered=self.window.video.stop,
            enabled=False,
        )

        icon = self.style.standardIcon(
            QStyle.StandardPixmap.SP_MediaSkipBackward
        )
        self.previous_frame_action = QAction(
            icon,
            "Go to the previous frame",
            self.window,
            shortcut=Qt.Key.Key_Left,
            triggered=partial(self.window.video.move_to_frame, -1),
            enabled=False,
        )
        self.fifth_previous_frame_action = QAction(
            icon,
            "Go to the fifth previous frame",
            self.window,
            shortcut=QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Left),
            triggered=partial(self.window.video.move_to_frame, -5),
            enabled=False,
        )
        self.tenth_previous_frame_action = QAction(
            icon,
            "Go to the tenth previous frame",
            self.window,
            shortcut=QKeySequence(
                Qt.Modifier.CTRL | Qt.Modifier.SHIFT | Qt.Key.Key_Left
            ),
            triggered=partial(self.window.video.move_to_frame, -10),
            enabled=False,
        )

        icon = self.style.standardIcon(
            QStyle.StandardPixmap.SP_MediaSkipForward
        )
        self.next_frame_action = QAction(
            icon,
            "Go to the next frame",
            self.window,
            shortcut=Qt.Key.Key_Right,
            triggered=partial(self.window.video.move_to_frame, 1),
            enabled=False,
        )
        self.fifth_next_frame_action = QAction(
            icon,
            "Go to the fifth next frame",
            self.window,
            shortcut=QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Right),
            triggered=partial(self.window.video.move_to_frame, 5),
            enabled=False,
        )
        self.tenth_next_frame_action = QAction(
            icon,
            "Go to the tenth next frame",
            self.window,
            shortcut=QKeySequence(
                Qt.Modifier.CTRL | Qt.Modifier.SHIFT | Qt.Key.Key_Right
            ),
            triggered=partial(self.window.video.move_to_frame, 10),
            enabled=False,
        )

        for action in [
            self.play_action,
            self.stop_action,
            self.tenth_previous_frame_action,
            self.fifth_previous_frame_action,
            self.previous_frame_action,
            self.next_frame_action,
            self.fifth_next_frame_action,
            self.tenth_next_frame_action,
        ]:
            menu.addAction(action)

    def create_edit_menu(self):
        menu = self.menu_bar.addMenu("&Edit")

        # Edit Timeline submenu
        edit_timeline_menu = menu.addMenu("&Timeline")

        self.add_timeline_action = QAction(
            "Add Timeline",
            self.window,
            triggered=self.window.time_pane.handle_timeline,
            enabled=False,
        )
        self.add_timeline_action.setShortcuts(
            [
                QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Return),
                QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Enter),
            ]
        )
        edit_timeline_menu.addAction(self.add_timeline_action)

        select_next_timeline_action = QAction(
            "Select Next Timeline",
            self.window,
            enabled=True,
        )
        select_next_timeline_action.triggered.connect(
            partial(self.select_cycle_timeline, 1)
        )
        select_next_timeline_action.setShortcuts([Qt.Key.Key_Down])
        edit_timeline_menu.addAction(select_next_timeline_action)

        select_previous_timeline_action = QAction(
            "Select Previous Timeline",
            self.window,
            enabled=True,
        )
        select_previous_timeline_action.triggered.connect(
            partial(self.select_cycle_timeline, -1)
        )
        select_previous_timeline_action.setShortcuts([Qt.Key.Key_Up])
        edit_timeline_menu.addAction(select_previous_timeline_action)

        # Edit Occurrence submenu
        edit_occurrence_menu = menu.addMenu("&Occurrence")

        self.add_occurrence_action = QAction(
            "Start Occurrence",
            self.window,
            triggered=self.add_occurrence,
            enabled=False,
        )
        self.add_occurrence_action.setShortcuts(
            [Qt.Key.Key_Return, Qt.Key.Key_Enter]
        )
        edit_occurrence_menu.addAction(self.add_occurrence_action)

        self.abort_occurrence_creation_action = QAction(
            "Abort Current Occurrence",
            self.window,
            triggered=self.window.time_pane.abort_occurrence_creation,
            enabled=False,
        )
        self.abort_occurrence_creation_action.setShortcuts([Qt.Key.Key_Escape])
        edit_occurrence_menu.addAction(self.abort_occurrence_creation_action)

        delete_occurrence_action = QAction(
            "Delete Occurrence",
            self.window,
            triggered=self.window.time_pane.delete_occurrence,
            enabled=True,
        )
        delete_occurrence_action.setShortcuts(
            [Qt.Key.Key_Backspace, Qt.Key.Key_Delete]
        )
        edit_occurrence_menu.addAction(delete_occurrence_action)

    def add_occurrence(self):
        self.window.time_pane.handle_occurrence()

    def select_cycle_timeline(self, delta):
        self.window.time_pane.select_cycle_timeline(delta)

    def create_view_menu(self):
        # Add actions to view menu
        menu = self.menu_bar.addMenu("&View")
        self.fullscreen_action = QAction(
            self.style.standardIcon(QStyle.StandardPixmap.SP_TitleBarMaxButton),
            "Toggle Fullscreen",
            self.window,
            shortcut=Qt.Key.Key_F11,
            triggered=self.window.on_fullscreen,
        )

        menu.addAction(self.fullscreen_action)

    def create_help_menu(self):
        # Help menu
        menu = self.menu_bar.addMenu("&Help")

        self.about_action = QAction(
            "About ViCodePy",
            self.window,
            triggered=self.about,
        )
        menu.addAction(self.about_action)

        self.visit_pypi_action = QAction(
            "Visit PyPI project",
            self.window,
            triggered=open_pypi_url,
        )
        menu.addAction(self.visit_pypi_action)

        self.visit_repository_action = QAction(
            "Visit Git repository",
            self.window,
            triggered=open_repository_url,
        )
        menu.addAction(self.visit_repository_action)

    def about(self):
        About().exec()

    def start_occurence(self):
        self.start_finish_occurrence("Start occurrence", False)

    def finish_occurence(self):
        self.start_finish_occurrence("Finish occurrence", True)

    def start_finish_occurrence(self, text, enable):
        self.add_occurrence_action.setText(text)
        self.abort_occurrence_creation_action.setEnabled(enable)
